#!/usr/bin/env python
__author__      = "Olman Quiros Jimenez"
__copyright__   = "Copyright 2024"
__license__     = "MIT"


class LibS3Error(Exception):
    """Base class for libs3 exceptions"""
    pass